@@ -1,15 +1,16 @@
 /* cmd line facility for slang */
-/* Copyright (c) 1992, 1995 John E. Davis
- * All rights reserved.
- * 
+/* Copyright (c) 1992, 1998 John E. Davis
+ * This file is part of the S-Lang library.
+ *
  * You may distribute under the terms of either the GNU General Public
  * License or the Perl Artistic License.
  */
 #include "config.h"
+#include "sl-feat.h"
 
 #include <stdio.h>
 
-#ifdef FLOAT_TYPE
+#if SLANG_HAS_FLOAT
 # include <math.h>
 #endif
 
@@ -23,7 +24,6 @@
 extern double atof ();
 #endif
 
-
 static SLcmd_Cmd_Type *SLcmd_find_command (char *s, SLcmd_Cmd_Type *cmd)
 {
    char *cmdstr;
@@ -39,57 +39,136 @@ static SLcmd_Cmd_Type *SLcmd_find_command (char *s, SLcmd_Cmd_Type *cmd)
    return NULL;
 }
 
+static int extract_token (char **strptr, char *buf)
+{
+   char *s, *b;
+   char ch, quote;
+
+   *buf = 0;
+
+   s = *strptr;
+   while (((ch = *s) != 0)
+	  && ((ch == ' ') || (ch == '\t') || (ch == '\n')))
+     s++;
+
+   *strptr = s;
+
+   if (ch == 0) return 0;
+   if (ch == '%') return 0;
+
+   b = buf;
+
+   *b++ = ch;
+   s++;
+
+   if ((ch == '\'') || (ch == '"'))
+     {
+	quote = ch;
+	while ((ch = *s) != 0)
+	  {
+	     s++;
+	     *b++ = ch;
+	     if (ch == quote)
+	       break;
+
+	     if (ch == '\\')
+	       {
+		  if (0 == (ch = *s))
+		    break;
+		  *b++ = ch;
+		  s++;
+	       }
+	  }
+	*strptr = s;
+	*b = 0;
+	return 1;
+     }
+
+   while (((ch = *s) != 0)
+	  && (ch != ' ')
+	  && (ch != '\t')
+	  && (ch != '\n')
+	  && (ch != '%'))
+     *b++ = *s++;
+
+   *strptr = s;
+   *b = 0;
+   return 1;
+}
+
 int SLcmd_execute_string (char *str, SLcmd_Cmd_Table_Type *table)
 {
    char *s, *b = NULL, *arg_type, *last_str, *cmd_name;
    SLcmd_Cmd_Type *cmd;
-   char buf[256];
+   char buf[1024];
    int token_present;
-   int i, ret = -1;
+   int i;
+   int status;
    unsigned int len;
    int argc;
-   
-   if ((0 == SLang_extract_token (&str, buf, 0))
-       || (*buf == '%'))
-     return 0;
-   if (SLang_Error) return -1;
-   
+
+   if (strlen (str) >= sizeof (buf))
+     {
+	SLang_doerror ("Input line too long");
+	return -1;
+     }
+
+   status = extract_token (&str, buf);
+   if (status <= 0) return status;
+
    if (((len = strlen (buf)) >= 32)
        || (NULL == (cmd = SLcmd_find_command (buf, table->table))))
      {
-	_SLdo_error ("%s: invalid command.", buf);
+	SLang_verror (SL_UNDEFINED_NAME,"%s: invalid command", buf);
 	return -1;
      }
 
    if (NULL == (cmd_name = SLmake_string (buf)))
      return -1;
-   
+
    argc = 0;
    table->string_args[argc++] = cmd_name;
-   
+
    arg_type = cmd->arg_type;
-   
+
+   status = -1;
    while (*arg_type)
      {
 	int guess_type = 0;
-	
+
 	last_str = str;
-	token_present = SLang_extract_token (&str, buf, 0);
-	if (SLang_Error) goto error;
+
+	if (-1 == (token_present = extract_token (&str, buf)))
+	  goto error;
+
 	if (token_present)
 	  {
 	     b = buf;
 	     len = strlen (b);
+
 	     if ((*b == '"') && (len > 1))
 	       {
 		  b++;
 		  len -= 2;
 		  b[len] = 0;
-		  guess_type = STRING_TYPE;
+		  guess_type = SLANG_STRING_TYPE;
+		  SLexpand_escaped_string (buf, b, b + len);
+		  len = strlen (buf);
+	       }
+	     else if ((*b == '\'') && (len > 1))
+	       {
+		  char ch;
+		  b++;
+		  len -= 2;
+		  b[len] = 0;
+		  guess_type = SLANG_INT_TYPE;
+		  (void) _SLexpand_escaped_char (b, &ch);
+		  sprintf (buf, "%d", (unsigned char) ch);
+		  len = strlen (buf);
 	       }
 	     else guess_type = SLang_guess_type (buf);
 	  }
-	
+
 	switch (*arg_type++)
 	  {
 	     /* variable argument number */
@@ -98,64 +177,62 @@ int SLcmd_execute_string (char *str, SLcmd_Cmd_Table_Type *table)
 	   case 'V':
 	     if (token_present == 0)
 	       {
-		  _SLdo_error ("%s: Expecting argument", cmd_name);
+		  SLang_verror (SL_INVALID_PARM, "%s: Expecting argument", cmd_name);
 		  goto error;
 	       }
-	     
+
 	     while (*last_str == ' ') last_str++;
 	     len = strlen (last_str);
 	     str = last_str + len;
-	     
+
 	     s = SLmake_nstring (last_str, len);
 	     if (s == NULL) goto error;
-	     table->arg_type[argc] = STRING_TYPE;
+	     table->arg_type[argc] = SLANG_STRING_TYPE;
 	     table->string_args[argc++] = s;
 	     break;
-	     
+
 	   case 's':
 	     if (token_present == 0) break;
 	   case 'S':
 	     if (token_present == 0)
 	       {
-		  _SLdo_error ("%s: Expecting string argument", cmd_name);
+		  SLang_verror (SL_TYPE_MISMATCH, "%s: Expecting string argument", cmd_name);
 		  goto error;
 	       }
-	     
-	     s = SLmake_nstring (b, len);
+
+	     s = SLmake_nstring (buf, len);
 
 	     if (s == NULL) goto error;
-	     table->arg_type[argc] = STRING_TYPE;
+	     table->arg_type[argc] = SLANG_STRING_TYPE;
 	     table->string_args[argc++] = s;
 	     break;
-	     
+
 	     /* integer argument */
 	   case 'i':
 	     if (token_present == 0) break;
 	   case 'I':
-	     if ((token_present == 0) || (INT_TYPE != guess_type))
+	     if ((token_present == 0) || (SLANG_INT_TYPE != guess_type))
 	       {
-		  _SLdo_error ("%s: Expecting integer argument", cmd_name);
-		  SLang_Error = TYPE_MISMATCH;
+		  SLang_verror (SL_TYPE_MISMATCH, "%s: Expecting integer argument", cmd_name);
 		  goto error;
 	       }
 
-	     table->arg_type[argc] = INT_TYPE;
+	     table->arg_type[argc] = SLANG_INT_TYPE;
 	     table->int_args[argc++] = SLatoi((unsigned char *) buf);
 	     break;
 
 	     /* floating point arg */
-#ifdef FLOAT_TYPE
+#if SLANG_HAS_FLOAT
 	   case 'f':
 	     if (token_present == 0) break;
 	   case 'F':
-	     if ((token_present == 0) || (STRING_TYPE == guess_type))
+	     if ((token_present == 0) || (SLANG_STRING_TYPE == guess_type))
 	       {
-		  _SLdo_error ("%s: Expecting float argument", cmd_name);
-		  SLang_Error = TYPE_MISMATCH;
+		  SLang_verror (SL_TYPE_MISMATCH, "%s: Expecting double argument", cmd_name);
 		  goto error;
 	       }
-	     table->arg_type[argc] = FLOAT_TYPE;
-	     table->float_args[argc++] = atof(buf);
+	     table->arg_type[argc] = SLANG_DOUBLE_TYPE;
+	     table->double_args[argc++] = atof(buf);
 	     break;
 #endif
 	     /* Generic type */
@@ -164,52 +241,47 @@ int SLcmd_execute_string (char *str, SLcmd_Cmd_Table_Type *table)
 	   case 'G':
 	     if (token_present == 0)
 	       {
-		  _SLdo_error ("%s: Expecting argument", cmd_name);
-		  SLang_Error = TYPE_MISMATCH;
+		  SLang_verror (SL_TYPE_MISMATCH, "%s: Expecting argument", cmd_name);
 		  goto error;
 	       }
-	     
+
 	     switch (guess_type)
 	       {
-		case INT_TYPE:
-		  table->arg_type[argc] = INT_TYPE;
+		case SLANG_INT_TYPE:
+		  table->arg_type[argc] = SLANG_INT_TYPE;
 		  table->int_args[argc++] = SLatoi((unsigned char *) buf);
 		  break;
-		  
-		case STRING_TYPE:
-		  s = SLmake_nstring (b, len);
+
+		case SLANG_STRING_TYPE:
+		  s = SLmake_nstring (buf, len);
 		  if (s == NULL) goto error;
-		  
-		  table->arg_type[argc] = STRING_TYPE;
+
+		  table->arg_type[argc] = SLANG_STRING_TYPE;
 		  table->string_args[argc++] = s;
 		  break;
-#ifdef FLOAT_TYPE
-		case FLOAT_TYPE:
-		  table->arg_type[argc] = FLOAT_TYPE;
-		  table->float_args[argc++] = atof(buf);
+#if SLANG_HAS_FLOAT
+		case SLANG_DOUBLE_TYPE:
+		  table->arg_type[argc] = SLANG_DOUBLE_TYPE;
+		  table->double_args[argc++] = atof(buf);
 #endif
 	       }
 	     break;
 	  }
      }
-   
+
    /*                 call function */
-   ret = (*cmd->cmdfun)(argc, table);
-   
+   status = (*cmd->cmdfun)(argc, table);
+
    error:
    for (i = 0; i < argc; i++)
      {
 	if (NULL != table->string_args[i])
 	  {
-	     SLFREE (table->string_args[i]);
+	     SLfree (table->string_args[i]);
 	     table->string_args[i] = NULL;
 	  }
      }
-   return ret;
-   
+   return status;
+
 }
 
-   
-
-
-   
