@@ -1,22 +1,21 @@
 /* Keymap routines for SLang.  The role of these keymap routines is simple:
- * Just read keys from the tty and return a pointer to a keymap structure.  
- * That is, a keymap is simple a mapping of strings (keys from tty) to 
- * structures.  Also included are routines for managing the keymaps. 
+ * Just read keys from the tty and return a pointer to a keymap structure.
+ * That is, a keymap is simple a mapping of strings (keys from tty) to
+ * structures.  Also included are routines for managing the keymaps.
  */
-/* Copyright (c) 1992, 1995 John E. Davis
- * All rights reserved.
- * 
+/* Copyright (c) 1992, 1998 John E. Davis
+ * This file is part of the S-Lang library.
+ *
  * You may distribute under the terms of either the GNU General Public
  * License or the Perl Artistic License.
  */
 
-
 #include "config.h"
+#include "sl-feat.h"
 
 #include <stdio.h>
 #include <string.h>
 
-
 #ifdef HAVE_STDLIB_H
 # include <stdlib.h>
 #endif
@@ -36,25 +35,20 @@ int SLang_Key_TimeOut_Flag = 0;	       /* true if more than 1 sec has elapsed
 int SLang_Last_Key_Char;
 
 
-
-
-SLKeyMap_List_Type SLKeyMap_List[SLANG_MAX_KEYMAPS];   /* these better be inited to 0! */
+SLKeyMap_List_Type SLKeyMap_List[SLANG_MAX_KEYMAPS];
 
 static SLang_Key_Type *malloc_key(unsigned char *str)
 {
    SLang_Key_Type *neew;
 
-   if (NULL == (neew = (SLang_Key_Type *) SLMALLOC(sizeof(SLang_Key_Type))))
-     {
-	SLang_Error = SL_MALLOC_ERROR;
-	return NULL;
-     }
+   if (NULL == (neew = (SLang_Key_Type *) SLmalloc(sizeof(SLang_Key_Type))))
+     return NULL;
+
    SLMEMSET ((char *) neew, 0, sizeof (SLang_Key_Type));
    SLMEMCPY((char *) neew->str, (char *) str, (unsigned int) *str);
    return(neew);
 }
 
-
 static SLKeyMap_List_Type *add_keymap(char *name, SLang_Key_Type *map)
 {
    int i;
@@ -65,33 +59,24 @@ static SLKeyMap_List_Type *add_keymap(char *name, SLang_Key_Type *map)
 	  {
 	     SLKeyMap_List[i].keymap = map;
 	     strncpy((char *) SLKeyMap_List[i].name, (char *) name, MAX_KEYMAP_NAME_LEN);
-	     SLKeyMap_List[i].name[MAX_KEYMAP_NAME_LEN] = 0; 
+	     SLKeyMap_List[i].name[MAX_KEYMAP_NAME_LEN] = 0;
 	     return &SLKeyMap_List[i];
 	  }
      }
-   SLang_Error = UNKNOWN_ERROR;
+   SLang_Error = SL_UNKNOWN_ERROR;
    /* SLang_doerror ("Keymap quota exceeded."); */
    return NULL;
 }
 
-#ifdef SLKEYMAP_OBSOLETE
-VOID_STAR 
-#else
-FVOID_STAR
-#endif
-SLang_find_key_function(char *name, SLKeyMap_List_Type *keymap)
+FVOID_STAR SLang_find_key_function(char *name, SLKeyMap_List_Type *keymap)
 {
    SLKeymap_Function_Type *fp = keymap -> functions;
-   
+
    while ((fp != NULL) && (fp->name != NULL))
      {
-	if ((*fp->name == *name) && !strcmp(fp->name, name)) 
+	if ((*fp->name == *name) && !strcmp(fp->name, name))
 	  {
-#ifdef SLKEYMAP_OBSOLETE
-	     return (VOID_STAR) fp->f;
-#else
 	     return (FVOID_STAR) fp->f;
-#endif
 	  }
 	fp++;
      }
@@ -106,28 +91,28 @@ static char *process_termcap_string (char *s, char *str, int *ip, int imax)
 {
    char c[3], *val;
    int i;
-   
+
    if ((0 == (c[0] = s[0]))
        || (0 == (c[1] = s[1]))
        || (s[2] != ')'))
      {
-	_SLdo_error ("setkey: ^(%s is badly formed.", s);
+	SLang_verror (SL_SYNTAX_ERROR, "setkey: ^(%s is badly formed", s);
 	return NULL;
      }
    s += 3;
-   
+
    c[2] = 0;
    if ((NULL == (val = SLtt_tgetstr (c)))
        || (*val == 0))
      return NULL;
-   
+
    i = *ip;
    while ((i < imax) && (*val != 0))
      {
 	str[i++] = *val++;
      }
    *ip = i;
-   
+
    return s;
 }
 #endif
@@ -169,7 +154,7 @@ char *SLang_process_keystring(char *s)
 	     ch = UPPER_CASE_KEY(ch);
 	     if (ch == '?') ch = 127; else ch = ch - 'A' + 1;
 	  }
-	
+
 	if (i >= 32) break;
 	str[i++] = ch;
      }
@@ -177,22 +162,21 @@ char *SLang_process_keystring(char *s)
    return(str);
 }
 
-
 static int key_string_compare (unsigned char *a, unsigned char *b, unsigned int len)
 {
    unsigned char *amax = a + len;
    int cha, chb, cha_up, chb_up;
-   
+
    while (a < amax)
      {
 	cha = *a++;
 	chb = *b++;
-	
+
 	if (cha == chb) continue;
-	
+
 	cha_up = UPPER_CASE_KEY(cha);
 	chb_up = UPPER_CASE_KEY(chb);
-	
+
 	if (cha_up == chb_up)
 	  {
 	     /* Use case-sensitive result. */
@@ -204,7 +188,6 @@ static int key_string_compare (unsigned char *a, unsigned char *b, unsigned int
    return 0;
 }
 
-	    
 static char *Define_Key_Error = "Inconsistency in define key.";
 
 /* This function also performs an insertion in an ordered way. */
@@ -216,9 +199,9 @@ static int find_the_key (char *s, SLKeyMap_List_Type *kml, SLang_Key_Type **keyp
    unsigned char *str;
 
    *keyp = NULL;
-   
+
    str = (unsigned char *) SLang_process_keystring(s);
-   
+
    if (1 == (str_len = str[0]))
      return 0;
 
@@ -232,16 +215,13 @@ static int find_the_key (char *s, SLKeyMap_List_Type *kml, SLang_Key_Type **keyp
 	     SLang_doerror (Define_Key_Error);
 	     return -2;
 	  }
-	
-	/* copy keymap uses the pointers so do not free these if copied */
-	if ((SLKeyMap_List[1].keymap == NULL)
-	    && (key->type == SLKEY_F_INTERPRET))
-	  {
-	     SLFREE (key->f.s);
-	  }
+
+	if (key->type == SLKEY_F_INTERPRET)
+	  SLang_free_slstring (key->f.s);
+
 	key->str[0] = str_len;
 	key->str[1] = ch;
-	
+
 	*keyp = key;
 	return 0;
      }
@@ -251,91 +231,69 @@ static int find_the_key (char *s, SLKeyMap_List_Type *kml, SLang_Key_Type **keyp
      {
 	int cmp;
 	unsigned int key_len, len;
-	
+
 	last = key;
 	key = key->next;
-	
+
 	if ((key != NULL) && (key->str != NULL))
 	  {
 	     len = key_len = key->str[0];
 	     if (len > str_len) len = str_len;
-	     
+
 	     cmp = key_string_compare (str + 1, key->str + 1, len - 1);
 
 	     if (cmp > 0)
 	       continue;
-	     
+
 	     if (cmp == 0)
 	       {
 		  if (key_len != str_len)
 		    {
-		       SLang_doerror (Define_Key_Error); 
+		       SLang_doerror (Define_Key_Error);
 		       return -2;
 		    }
 
-		  /* copy keymap uses the pointers so do not free these if copied */
-		  if ((SLKeyMap_List[1].keymap == NULL)
-		      && (key->type == SLKEY_F_INTERPRET))
-		    {
-#ifdef SLKEYMAP_OBSOLETE	
-		       SLFREE(key->f);
-#else
-		       SLFREE(key->f.s);
-#endif
-		    }
-		  
+		  if (key->type == SLKEY_F_INTERPRET)
+		    SLang_free_slstring (key->f.s);
+
 		  *keyp = key;
 		  return 0;
 	       }
 	     /* Drop to cmp < 0 case */
 	  }
-	
+
 	if (NULL == (neew = malloc_key(str))) return -1;
-	     
+
 	neew -> next = key;
 	last -> next = neew;
-	
+
 	*keyp = neew;
 	return 0;
      }
 }
 
 /* returns -2 if inconsistent, -1 if malloc error, 0 upon success */
-#ifdef SLKEYMAP_OBSOLETE
-int SLang_define_key1 (char *s, VOID_STAR f, unsigned int type, SLKeyMap_List_Type *kml)
-#else
 int SLkm_define_key (char *s, FVOID_STAR f, SLKeyMap_List_Type *kml)
-#endif
 {
    SLang_Key_Type *key;
-#ifndef SLKEYMAP_OBSOLETE
    unsigned int type = SLKEY_F_INTRINSIC;
-#endif
    int ret;
-   
+
    ret = find_the_key (s, kml, &key);
    if ((ret != 0) || (key == NULL))
      return ret;
 
    key->type = type;
-#ifdef SLKEYMAP_OBSOLETE
-   key->f = f;
-#else
    key->f.f = f;
-#endif
    return 0;
 }
 
-int SLang_define_key(char *s, char *funct, SLKeyMap_List_Type *kml)
+int SLang_define_key (char *s, char *funct, SLKeyMap_List_Type *kml)
 {
    SLang_Key_Type *key;
-#ifdef SLKEYMAP_OBSOLETE
-   VOID_STAR f;
-#else
    FVOID_STAR f;
-#endif
    int ret;
-   
+
    ret = find_the_key (s, kml, &key);
    if ((ret != 0) || (key == NULL))
      return ret;
@@ -344,33 +302,24 @@ int SLang_define_key(char *s, char *funct, SLKeyMap_List_Type *kml)
 
    if (f == NULL)                      /* assume interpreted */
      {
-	char *str = SLmake_string (funct);
+	char *str = SLang_create_slstring (funct);
 	if (str == NULL) return -1;
 	key->type = SLKEY_F_INTERPRET;
-#ifdef SLKEYMAP_OBSOLETE
-	key->f = (VOID_STAR) str;
-#else
 	key->f.s = str;
-#endif
      }
-   else 
+   else
      {
 	key->type = SLKEY_F_INTRINSIC;
-#ifdef SLKEYMAP_OBSOLETE
-	key->f = f;
-#else
 	key->f.f = f;
-#endif
      }
    return 0;
 }
 
-#ifndef SLKEYMAP_OBSOLETE
 int SLkm_define_keysym (char *s, unsigned int keysym, SLKeyMap_List_Type *kml)
 {
    SLang_Key_Type *key;
    int ret;
-   
+
    ret = find_the_key (s, kml, &key);
 
    if ((ret != 0) || (key == NULL))
@@ -380,7 +329,6 @@ int SLkm_define_keysym (char *s, unsigned int keysym, SLKeyMap_List_Type *kml)
    key->f.keysym = keysym;
    return 0;
 }
-#endif
 
 SLang_Key_Type *SLang_do_key(SLKeyMap_List_Type *kml, int (*getkey)(void))
 {
@@ -389,51 +337,51 @@ SLang_Key_Type *SLang_do_key(SLKeyMap_List_Type *kml, int (*getkey)(void))
    unsigned char input_ch;
    register unsigned char chup, chlow;
    unsigned char key_ch = 0;
-   
+
    SLang_Last_Key_Char = (*getkey)();
    SLang_Key_TimeOut_Flag = 0;
-   
+
    if (SLANG_GETKEY_ERROR == (unsigned int) SLang_Last_Key_Char)
      return NULL;
-   
+
    input_ch = (unsigned char) SLang_Last_Key_Char;
-   
+
    key = (SLang_Key_Type *) &((kml->keymap)[input_ch]);
 
    /* if the next one is null, then we know this MAY be it. */
-   while (key->next == NULL) 
+   while (key->next == NULL)
      {
 	if (key->type != 0)
 	  return key;
-	
+
 	/* Try its opposite case counterpart */
 	chlow = LOWER_CASE_KEY(input_ch);
 	if (input_ch == chlow)
 	  input_ch = UPPER_CASE_KEY(input_ch);
-	
+
 	key = kml->keymap + input_ch;
 	if (key->type == 0)
 	  return NULL;
      }
-   
+
    /* It appears to be a prefix character in a key sequence. */
-   
+
    len = 1;			       /* already read one character */
    key = key->next;		       /* Now we are in the key list */
    kmax = NULL;			       /* set to end of list */
-   
+
    while (1)
-     {	
-	SLang_Key_TimeOut_Flag = 1;   
+     {
+	SLang_Key_TimeOut_Flag = 1;
 	SLang_Last_Key_Char = (*getkey)();
 	SLang_Key_TimeOut_Flag = 0;
-	
+
 	len++;
-	
+
 	if ((SLANG_GETKEY_ERROR == (unsigned int) SLang_Last_Key_Char)
 	    || SLKeyBoard_Quit)
 	  break;
-	
+
 	input_ch = (unsigned char) SLang_Last_Key_Char;
 
 	chup = UPPER_CASE_KEY(input_ch); chlow = LOWER_CASE_KEY(input_ch);
@@ -448,11 +396,11 @@ SLang_Key_Type *SLang_do_key(SLKeyMap_List_Type *kml, int (*getkey)(void))
 	       }
 	     key = key->next;
 	  }
-	
+
 	if (key == kmax) break;
 
 	/* If the input character is lowercase, check to see if there is
-	 * a lowercase match.  If so, set key to it.  Note: the 
+	 * a lowercase match.  If so, set key to it.  Note: the
 	 * algorithm assumes the sorting performed by key_string_compare.
 	 */
 	if (input_ch != key_ch)
@@ -474,17 +422,17 @@ SLang_Key_Type *SLang_do_key(SLKeyMap_List_Type *kml, int (*getkey)(void))
 		  next = next->next;
 	       }
 	  }
-	
+
 	/* Ok, we found the first position of a possible match.  If it
 	 * is exact, we are done.
 	 */
 	if ((unsigned int) key->str[0] == len + 1)
 	  return key;
-	
+
 	/* Apparantly, there are some ambiguities. Read next key to resolve
 	 * the ambiguity.  Adjust kmax to encompass ambiguities.
 	 */
-	
+
 	next = key->next;
 	while (next != kmax)
 	  {
@@ -502,8 +450,6 @@ SLang_Key_Type *SLang_do_key(SLKeyMap_List_Type *kml, int (*getkey)(void))
    return NULL;
 }
 
-
-
 void SLang_undefine_key(char *s, SLKeyMap_List_Type *kml)
 {
    int n, i;
@@ -515,25 +461,19 @@ void SLang_undefine_key(char *s, SLKeyMap_List_Type *kml)
 
    if (0 == (n = *str++ - 1)) return;
    i = *str;
-   
+
    last = key_root = (SLang_Key_Type *) &(keymap[i]);
    key = key_root->next;
-   
+
    while (key != NULL)
      {
 	next = key->next;
 	if (0 == SLMEMCMP ((char *)(key->str + 1), (char *) str, n))
 	  {
-	     if ((SLKeyMap_List[1].keymap == NULL)
-		 && (key->type == SLKEY_F_INTERPRET))
-	       {
-#ifdef SLKEYMAP_OBSOLETE
-		  SLFREE(key->f);
-#else
-		  SLFREE (key->f.s);
-#endif
-	       }
-	     SLFREE(key);
+	     if (key->type == SLKEY_F_INTERPRET)
+	       SLang_free_slstring (key->f.s);
+
+	     SLfree((char *) key);
 	     last->next = next;
 	  }
 	else last = key;
@@ -543,11 +483,7 @@ void SLang_undefine_key(char *s, SLKeyMap_List_Type *kml)
    if (n == 1)
      {
 	*key_root->str = 0;
-#ifdef SLKEYMAP_OBSOLETE
-	key_root->f = NULL;
-#else
 	key_root->f.f = NULL;
-#endif
 	key_root->type = 0;
      }
 }
@@ -574,40 +510,27 @@ char *SLang_make_keystring(unsigned char *s)
    return(buf);
 }
 
-
-
-
 static SLang_Key_Type *copy_keymap(SLKeyMap_List_Type *kml)
 {
    int i;
    SLang_Key_Type *neew, *old, *new_root, *km;
 
-   if (NULL == (new_root = (SLang_Key_Type *) SLCALLOC(256, sizeof(SLang_Key_Type))))
-     {
-	SLang_Error = SL_MALLOC_ERROR;
-	return NULL;
-     }
+   if (NULL == (new_root = (SLang_Key_Type *) SLcalloc(256, sizeof(SLang_Key_Type))))
+     return NULL;
 
    if (kml == NULL) return new_root;
    km = kml->keymap;
 
-   
    for (i = 0; i < 256; i++)
      {
 	old = &(km[i]);
 	neew = &(new_root[i]);
-#ifdef SLKEYMAP_OBSOLETE
-	neew->f = old->f;
-#else
-#if 0
+
 	if (old->type == SLKEY_F_INTERPRET)
-	  neew->f.s = old->f.s;
-	else 
+	  neew->f.s = SLang_create_slstring (old->f.s);
+	else
 	  neew->f.f = old->f.f;
-#else
-	SLMEMCPY((char *) &neew->f, (char *) &old->f, sizeof (neew->f));
-#endif
-#endif
+
 	neew->type = old->type;
 	SLMEMCPY((char *) neew->str, (char *) old->str, (unsigned int) *old->str);
 
@@ -616,18 +539,12 @@ static SLang_Key_Type *copy_keymap(SLKeyMap_List_Type *kml)
 	  {
 	     neew->next = malloc_key((unsigned char *) old->str);
 	     neew = neew->next;
-#ifdef SLKEYMAP_OBSOLETE
-	     neew->f = old->f;
-#else
-# if 0
+
 	     if (old->type == SLKEY_F_INTERPRET)
-	       neew->f.s = old->f.s;
+	       neew->f.s = SLang_create_slstring (old->f.s);
 	     else
 	       neew->f.f = old->f.f;
-# else
-	     SLMEMCPY((char *) &neew->f, (char *) &old->f, sizeof (neew->f));
-# endif
-#endif
+
 	     neew->type = old->type;
 	     old = old->next;
 	  }
@@ -636,17 +553,16 @@ static SLang_Key_Type *copy_keymap(SLKeyMap_List_Type *kml)
    return(new_root);
 }
 
-
 SLKeyMap_List_Type *SLang_create_keymap(char *name, SLKeyMap_List_Type *map)
 {
    SLang_Key_Type *neew;
    SLKeyMap_List_Type *new_map;
-   
+
    if ((NULL == (neew = copy_keymap(map)))
        || (NULL == (new_map = add_keymap(name, neew)))) return NULL;
-   
+
    if (map != NULL) new_map -> functions = map -> functions;
-   
+
    return new_map;
 }
 
