@@ -1,17 +1,16 @@
 /* SLang_read_line interface --- uses SLang tty stuff */
-/* Copyright (c) 1992, 1995 John E. Davis
- * All rights reserved.
- * 
+/* Copyright (c) 1992, 1998 John E. Davis
+ * This file is part of the S-Lang library.
+ *
  * You may distribute under the terms of either the GNU General Public
  * License or the Perl Artistic License.
  */
 
-
 #include "config.h"
+#include "sl-feat.h"
 
 #include <stdio.h>
 
-
 #include "slang.h"
 #include "_slang.h"
 
@@ -58,40 +57,40 @@ static int rl_right (void)
 static int rl_left (void)
 {
    if (This_RLI->point == 0) return 0;
-   This_RLI->point--; 
+   This_RLI->point--;
    return 1;
 }
 
 static int rl_self_insert (void)
 {
    unsigned char *pmin, *p;
-   
+
    if (This_RLI->len == This_RLI->buf_len)
      {
 	rl_beep ();
 	return 0;
      }
-   
+
    pmin = This_RLI->buf + This_RLI->point;
    p = This_RLI->buf + This_RLI->len;
-   while (p > pmin) 
+   while (p > pmin)
      {
 	*p = *(p - 1);
 	p--;
      }
    *pmin = SLang_Last_Key_Char;
-   
+
    This_RLI->len++;
    This_RLI->point++;
    if ((This_RLI->curs_pos + 2 >= This_RLI->edit_width)
        || (This_RLI->tt_insert == NULL)
        || (Char_Widths[SLang_Last_Key_Char] != 1)) return 1;
-   
+
    (*This_RLI->tt_insert)((char) SLang_Last_Key_Char);
    /* update screen buf */
    p = This_RLI->old_upd + (This_RLI->len - 1);
    pmin = This_RLI->old_upd + (This_RLI->point - 1);
-   while (p > pmin) 
+   while (p > pmin)
      {
 	*p = *(p - 1);
 	p--;
@@ -104,38 +103,37 @@ int SLang_rline_insert (char *s)
 {
    unsigned char *pmin, *p;
    int n;
-   
+
    n = strlen (s);
-   if (n > This_RLI->buf_len - This_RLI->len) 
+   if (n > This_RLI->buf_len - This_RLI->len)
      n = This_RLI->buf_len - This_RLI->len;
-   
+
    if (n == 0) return 0;
-   
+
    pmin = This_RLI->buf + This_RLI->point;
    p = This_RLI->buf + (This_RLI->len - 1);
-   
+
    while (p >= pmin)
      {
 	*(p + n) = *p;
 	p--;
      }
    SLMEMCPY ((char *) pmin, s, n);
-   
+
    This_RLI->len += n;
    This_RLI->point += n;
    return n;
 }
 
-
 static int rl_deln (int n)
 {
    unsigned char *pmax, *p;
-   
+
    p = This_RLI->buf + This_RLI->point;
    pmax = This_RLI->buf + This_RLI->len;
-   
+
    if (p + n > pmax) n = (int) (pmax - p);
-   while (p < pmax) 
+   while (p < pmax)
      {
 	*p = *(p + n);
 	p++;
@@ -155,34 +153,33 @@ static int rl_quote_insert (void)
    SLang_Error = 0;
    SLang_Last_Key_Char = (*This_RLI->getkey)();
    rl_self_insert ();
-   if (SLang_Error == USER_BREAK) SLang_Error = 0;
+   if (SLang_Error == SL_USER_BREAK) SLang_Error = 0;
    else SLang_Error = err;
    return 1;
 }
 
-   
 static int rl_trim (void)
 {
    unsigned char *p, *pmax, *p1;
    p = This_RLI->buf + This_RLI->point;
    pmax = This_RLI->buf + This_RLI->len;
-   
-   if (p == pmax) 
+
+   if (p == pmax)
      {
 	if (p == This_RLI->buf) return 0;
 	p--;
      }
-   
+
    if ((*p != ' ') && (*p != '\t')) return 0;
    p1 = p;
    while ((p1 < pmax) && ((*p1 == ' ') || (*p1 == '\t'))) p1++;
    pmax = p1;
    p1 = This_RLI->buf;
-   
+
    while ((p >= p1) && ((*p == ' ') || (*p == '\t'))) p--;
    if (p == pmax) return 0;
    p++;
-   
+
    This_RLI->point = (int) (p - p1);
    return rl_deln ((int) (pmax - p));
 }
@@ -218,9 +215,9 @@ static int rl_enter (void)
 
 static SLKeyMap_List_Type *RL_Keymap;
 
-/* This update is designed for dumb terminals.  It assumes only that the 
+/* This update is designed for dumb terminals.  It assumes only that the
  * terminal can backspace via ^H, and move cursor to start of line via ^M.
- * There is a hook so the user can provide a more sophisticated update if 
+ * There is a hook so the user can provide a more sophisticated update if
  * necessary.
  */
 
@@ -228,14 +225,13 @@ static void position_cursor (int col)
 {
    unsigned char *p, *pmax;
    int dc;
-   
-   if (col == This_RLI->curs_pos) 
+
+   if (col == This_RLI->curs_pos)
      {
 	fflush (stdout);
 	return;
      }
-   
-   
+
    if (This_RLI->tt_goto_column != NULL)
      {
 	(*This_RLI->tt_goto_column)(col);
@@ -243,7 +239,7 @@ static void position_cursor (int col)
 	fflush (stdout);
 	return;
      }
-   
+
    dc = This_RLI->curs_pos - col;
    if (dc < 0)
      {
@@ -272,16 +268,15 @@ static void position_cursor (int col)
 static void erase_eol (SLang_RLine_Info_Type *rli)
 {
    unsigned char *p, *pmax;
-   
+
    p = rli->old_upd + rli->curs_pos;
    pmax = rli->old_upd + rli->old_upd_len;
-   
+
    while (p++ < pmax) putc(' ', stdout);
-	
+
    rli->curs_pos = rli->old_upd_len;
 }
 
-   
 static unsigned char *spit_out(SLang_RLine_Info_Type *rli, unsigned char *p)
 {
    unsigned char *pmax;
@@ -296,7 +291,7 @@ static void really_update (SLang_RLine_Info_Type *rli, int new_curs_position)
 {
    unsigned char *b = rli->old_upd, *p = rli->new_upd, chb, chp;
    unsigned char *pmax;
-   
+
    if (rli->update_hook != NULL)
      {
 	(*rli->update_hook)(p, rli->edit_width, new_curs_position);
@@ -308,7 +303,7 @@ static void really_update (SLang_RLine_Info_Type *rli, int new_curs_position)
 	  {
 	     chb = *b++; chp = *p++;
 	     if (chb == chp) continue;
-	     
+
 	     if (rli->old_upd_len <= rli->new_upd_len)
 	       {
 		  /* easy one */
@@ -321,9 +316,9 @@ static void really_update (SLang_RLine_Info_Type *rli, int new_curs_position)
 	  }
 	position_cursor (new_curs_position);
      }
-   
+
    /* update finished, so swap */
-   
+
    rli->old_upd_len = rli->new_upd_len;
    p = rli->old_upd;
    rli->old_upd = rli->new_upd;
@@ -336,18 +331,18 @@ static void RLupdate (SLang_RLine_Info_Type *rli)
    int want_cursor_pos;
    unsigned char *b, chb, *b_point, *p;
    int no_echo;
-   
+
    no_echo = rli->flags & SL_RLINE_NO_ECHO;
 
    b_point = (unsigned char *) (rli->buf + rli->point);
    *(rli->buf + rli->len) = 0;
-   
-   /* expand characters for output buffer --- handle prompt first.  
+
+   /* expand characters for output buffer --- handle prompt first.
     * Do two passes --- first to find out where to begin upon horiz
     * scroll and the second to actually fill the buffer. */
    len = 0;
    count = 2;			       /* once for prompt and once for buf */
-   
+
    b = (unsigned char *) rli->prompt;
    while (count--)
      {
@@ -372,7 +367,7 @@ static void RLupdate (SLang_RLine_Info_Type *rli)
 	b = (unsigned char *) rli->buf;
 	if (count == 1) want_cursor_pos = prompt_len = len;
      }
-   
+
    if (len < rli->edit_width - rli->dhscroll) start_len = 0;
    else if ((rli->start_column > len)
 	    || (rli->start_column + rli->edit_width <= len))
@@ -382,23 +377,22 @@ static void RLupdate (SLang_RLine_Info_Type *rli)
      }
    else start_len = rli->start_column;
    rli->start_column = start_len;
-   
+
    want_cursor_pos = len - start_len;
 
-   
    /* second pass */
    p = rli->new_upd;
-	
+
    len = 0;
    count = 2;
    b = (unsigned char *) rli->prompt;
    if (b == NULL) b = (unsigned char *) "";
-   
+
    while ((len < start_len) && (*b))
      {
 	len += Char_Widths[*b++];
      }
-   
+
    tw = 0;
    if (*b == 0)
      {
@@ -410,7 +404,7 @@ static void RLupdate (SLang_RLine_Info_Type *rli)
 	tw = rli->tab;
 	count--;
      }
-      
+
    len = 0;
    while (count--)
      {
@@ -421,7 +415,7 @@ static void RLupdate (SLang_RLine_Info_Type *rli)
 	  {
 	     dlen = Char_Widths[chb];
 	     if (dlen == 1) *p++ = chb;
-	     else 
+	     else
 	       {
 		  if ((chb == '\t') && tw)
 		    {
@@ -431,14 +425,14 @@ static void RLupdate (SLang_RLine_Info_Type *rli)
 		       while (dlen--) *p++ = ' ';
 		       dlen = 0;
 		    }
-		  else 
+		  else
 		    {
 		       if (dlen == 3)
 			 {
 			    chb &= 0x7F;
 			    *p++ = '~';
 			 }
-		  
+
 		       *p++ = '^';
 		       if (chb == 127)  *p++ = '?';
 		       else *p++ = chb + '@';
@@ -450,7 +444,7 @@ static void RLupdate (SLang_RLine_Info_Type *rli)
 	tw = rli->tab;
 	b = (unsigned char *) rli->buf;
      }
-   
+
    rli->new_upd_len = (int) (p - rli->new_upd);
    while (p < rli->new_upd + rli->edit_width) *p++ = ' ';
    really_update (rli, want_cursor_pos);
@@ -477,39 +471,111 @@ static int rl_eof_insert (void)
    return 0;
 }
 
+/* This is very naive.  It knows very little about nesting and nothing
+ * about quoting.
+ */
+static void blink_match (SLang_RLine_Info_Type *rli)
+{
+   unsigned char bra, ket;
+   unsigned int delta_column;
+   unsigned char *p, *pmin;
+   int dq_level, sq_level;
+   int level;
+
+   pmin = rli->buf;
+   p = pmin + rli->point;
+   if (pmin == p)
+     return;
+
+   ket = SLang_Last_Key_Char;
+   switch (ket)
+     {
+      case ')':
+	bra = '(';
+	break;
+      case ']':
+	bra = '[';
+	break;
+      case '}':
+	bra = '{';
+	break;
+      default:
+	return;
+     }
+
+   level = 0;
+   sq_level = dq_level = 0;
+
+   delta_column = 0;
+   while (p > pmin)
+     {
+	char ch;
+
+	p--;
+	delta_column++;
+	ch = *p;
+
+	if (ch == ket)
+	  {
+	     if ((dq_level == 0) && (sq_level == 0))
+	       level++;
+	  }
+	else if (ch == bra)
+	  {
+	     if ((dq_level != 0) || (sq_level != 0))
+	       continue;
+
+	     level--;
+	     if (level == 0)
+	       {
+		  rli->point -= delta_column;
+		  RLupdate (rli);
+		  (*rli->input_pending)(10);
+		  rli->point += delta_column;
+		  RLupdate (rli);
+		  break;
+	       }
+	     if (level < 0)
+	       break;
+	  }
+	else if (ch == '"') dq_level = !dq_level;
+	else if (ch == '\'') sq_level = !sq_level;
+     }
+}
+
 int SLang_read_line (SLang_RLine_Info_Type *rli)
 {
    unsigned char *p, *pmax;
    SLang_Key_Type *key;
-   
+
    SLang_Rline_Quit = 0;
    This_RLI = rli;
    p = rli->old_upd; pmax = p + rli->edit_width;
    while (p < pmax) *p++ = ' ';
-   
+
    /* Sanity checking */
    rli->len = strlen ((char *) rli->buf);
-   if (rli->len >= rli->buf_len) 
+   if (rli->len >= rli->buf_len)
      {
 	rli->len = 0;
 	*rli->buf = 0;
      }
    if (rli->point > rli->len) rli->point = rli->len;
    if (rli->point < 0) rli->point = 0;
-   
+
    rli->curs_pos = rli->start_column = 0;
    rli->new_upd_len = rli->old_upd_len = 0;
-   
+
    This_RLI->last_fun = NULL;
    if (rli->update_hook == NULL)
      putc ('\r', stdout);
 
    RLupdate (rli);
-   
+
    while (1)
      {
 	key = SLang_do_key (RL_Keymap, (int (*)(void)) rli->getkey);
-	
+
 	if ((key == NULL)
 #ifdef SLKEYMAP_OBSOLETE
 	    || (key->f == NULL)
@@ -526,21 +592,25 @@ int SLang_read_line (SLang_RLine_Info_Type *rli)
 		 && (*key->str == 2)
 		 && (This_RLI->len == 0))
 	       rl_eof_insert ();
-	     else if (key->type == SLKEY_F_INTRINSIC) 
+	     else if (key->type == SLKEY_F_INTRINSIC)
 	       {
 #ifdef SLKEYMAP_OBSOLETE
 		  if (((int (*)(void))(key->f))())
 		    RLupdate (rli);
 #else
-		  if ((key->f.f)()) 
+		  if ((key->f.f)())
 		    RLupdate (rli);
+
+		  if ((rli->flags & SL_RLINE_BLINK_MATCH)
+		      && (rli->input_pending != NULL))
+		    blink_match (rli);
 #endif
 	       }
-	
-	     if (SLang_Rline_Quit) 
+
+	     if (SLang_Rline_Quit)
 	       {
 		  This_RLI->buf[This_RLI->len] = 0;
-		  if (SLang_Error == USER_BREAK) 
+		  if (SLang_Error == SL_USER_BREAK)
 		    {
 		       SLang_Error = 0;
 		       return -1;
@@ -548,7 +618,7 @@ int SLang_read_line (SLang_RLine_Info_Type *rli)
 		  return This_RLI->len;
 	       }
 	  }
-	if (key != NULL) 
+	if (key != NULL)
 	  {
 #ifdef SLKEYMAP_OBSOLETE
 	     This_RLI->last_fun = key->f;
@@ -565,8 +635,6 @@ static int rl_abort (void)
    return rl_enter ();
 }
 
-
-
 /* TTY interface --- ANSI */
 
 static void ansi_goto_column (int n)
@@ -585,7 +653,7 @@ static int rl_next_line (void);
 static int rl_prev_line (void)
 {
    SLang_Read_Line_Type *prev;
-   
+
    if (((This_RLI->last_fun != (FVOID_STAR) rl_prev_line)
 	&& (This_RLI->last_fun != (FVOID_STAR) rl_next_line))
        || (This_RLI->last == NULL))
@@ -593,13 +661,13 @@ static int rl_prev_line (void)
 	prev = This_RLI->tail;
      }
    else prev = This_RLI->last->prev;
-   
+
    if (prev == NULL)
      {
 	rl_beep ();
 	return 0;
      }
-   
+
    rl_select_line (prev);
    return 1;
 }
@@ -612,7 +680,7 @@ static int rl_redraw (void)
 static int rl_next_line (void)
 {
    SLang_Read_Line_Type *next;
-   
+
    if (((This_RLI->last_fun != (FVOID_STAR) rl_prev_line)
 	&& (This_RLI->last_fun != (FVOID_STAR) rl_next_line))
        || (This_RLI->last == NULL))
@@ -620,9 +688,9 @@ static int rl_next_line (void)
 	 rl_beep ();
 	 return 0;
       }
-   
+
    next = This_RLI->last->next;
-   
+
    if (next == NULL)
      {
 	This_RLI->len = This_RLI->point = 0;
@@ -633,9 +701,7 @@ static int rl_next_line (void)
    return 1;
 }
 
-
-
-static SLKeymap_Function_Type SLReadLine_Functions[] = 
+static SLKeymap_Function_Type SLReadLine_Functions[] =
 {
    {"up", rl_prev_line},
    {"down", rl_next_line},
@@ -657,15 +723,15 @@ int SLang_init_readline (SLang_RLine_Info_Type *rli)
 {
    int ch;
    char simple[2];
-   
+
    if (RL_Keymap == NULL)
      {
 	simple[1] = 0;
 	if (NULL == (RL_Keymap = SLang_create_keymap ("ReadLine", NULL)))
 	  return -1;
-	
+
 	RL_Keymap->functions = SLReadLine_Functions;
-	
+
 	/* This breaks under some DEC ALPHA compilers (scary!) */
 #ifndef __DECC
 	for (ch = ' '; ch < 256; ch++)
@@ -704,8 +770,8 @@ int SLang_init_readline (SLang_RLine_Info_Type *rli)
 #else
 	SLkm_define_key (simple, (FVOID_STAR) rl_eof_insert, RL_Keymap);
 #endif
-	
-#ifndef pc_system
+
+#ifndef IBMPC_SYSTEM
 # ifdef SLKEYMAP_OBSOLETE
 	SLang_define_key1 ("^[[A", (VOID_STAR) rl_prev_line, SLKEY_F_INTRINSIC, RL_Keymap);
 	SLang_define_key1 ("^[[B", (VOID_STAR) rl_next_line, SLKEY_F_INTRINSIC, RL_Keymap);
@@ -753,7 +819,7 @@ int SLang_init_readline (SLang_RLine_Info_Type *rli)
 	SLang_define_key1 ("^P", (VOID_STAR) rl_prev_line, SLKEY_F_INTRINSIC, RL_Keymap);
 	SLang_define_key1 ("^N", (VOID_STAR) rl_next_line, SLKEY_F_INTRINSIC, RL_Keymap);
 	SLang_define_key1 ("^R", (VOID_STAR) rl_redraw, SLKEY_F_INTRINSIC, RL_Keymap);
-	
+
 	SLang_define_key1 ("`", (VOID_STAR) rl_quote_insert, SLKEY_F_INTRINSIC, RL_Keymap);
 	SLang_define_key1 ("\033\\", (VOID_STAR) rl_trim, SLKEY_F_INTRINSIC, RL_Keymap);
 #else
@@ -777,31 +843,31 @@ int SLang_init_readline (SLang_RLine_Info_Type *rli)
 	SLkm_define_key  ("^R", (FVOID_STAR) rl_redraw, RL_Keymap);
 	SLkm_define_key  ("`", (FVOID_STAR) rl_quote_insert, RL_Keymap);
 	SLkm_define_key  ("\033\\", (FVOID_STAR) rl_trim, RL_Keymap);
-#endif	
+#endif
 	if (SLang_Error) return -1;
      }
-   
+
    if (rli->prompt == NULL) rli->prompt = "";
    if (rli->keymap == NULL) rli->keymap = RL_Keymap;
    rli->old_upd = rli->upd_buf1;
    rli->new_upd = rli->upd_buf2;
-   *rli->buf = 0; 
+   *rli->buf = 0;
    rli->point = 0;
-   
+
    if (rli->flags & SL_RLINE_USE_ANSI)
      {
 	if (rli->tt_goto_column == NULL) rli->tt_goto_column = ansi_goto_column;
      }
-   
+
    if (Char_Widths[0] == 2) return 0;
-	
+
    for (ch = 0; ch < 32; ch++) Char_Widths[ch] = 2;
    for (ch = 32; ch < 256; ch++) Char_Widths[ch] = 1;
    Char_Widths[127] = 2;
-#ifndef pc_system
+#ifndef IBMPC_SYSTEM
    for (ch = 128; ch < 160; ch++) Char_Widths[ch] = 3;
 #endif
-   
+
    return 0;
 }
 
@@ -813,23 +879,23 @@ SLang_Read_Line_Type *SLang_rline_save_line (SLang_RLine_Info_Type *rli)
    if ((rli == NULL) || (rli->buf == NULL))
      return NULL;
 
-   if (NULL == (rl = (SLang_Read_Line_Type *) SLMALLOC (sizeof (SLang_Read_Line_Type)))
+   if (NULL == (rl = (SLang_Read_Line_Type *) SLmalloc (sizeof (SLang_Read_Line_Type)))
        || (NULL == (buf = (unsigned char *) SLmake_string ((char *)rli->buf))))
      {
-	if (rl != NULL) SLFREE (rl);
+	SLfree ((char *)rl);
 	return NULL;
      }
    rl->buf = buf;
    rl->buf_len = strlen ((char *)buf);
    rl->num = rl->misc = 0;
    rl->next = rl->prev = NULL;
-      
-   if (rli->tail != NULL) 
+
+   if (rli->tail != NULL)
      {
 	rli->tail->next = rl;
 	rl->prev = rli->tail;
      }
    rli->tail = rl;
-   
+
    return rl;
 }
