@@ -2,14 +2,15 @@
  * the slang SLtt interface.
  */
 
-/* Copyright (c) 1992, 1995 John E. Davis
- * All rights reserved.
- * 
+/* Copyright (c) 1992, 1998 John E. Davis
+ * This file is part of the S-Lang library.
+ *
  * You may distribute under the terms of either the GNU General Public
  * License or the Perl Artistic License.
  */
 
-#include <config.h>
+#include "config.h"
+#include "sl-feat.h"
 
 #include <stdio.h>
 #ifdef HAVE_STDLIB_H
@@ -23,7 +24,7 @@
  * The majority of the comments found in the file were taken from the
  * term(4) man page on an SGI.
  */
- 
+
 /* Short integers are stored in two 8-bit bytes.  The first byte contains
  * the least significant 8 bits of the value, and the second byte contains
  * the most significant 8 bits.  (Thus, the value represented is
@@ -52,7 +53,7 @@ static int make_integer (unsigned char *buf)
  * tic, and read by the routine setupterm [see curses(3X).]  The file is
  * divided into six parts in the following order:  the header, terminal
  * names, boolean flags, numbers, strings, and string table.
- * 
+ *
  * The header section begins the file.  This section contains six short
  * integers in the format described below.  These integers are (1) the magic
  * number (octal 0432); (2) the size, in bytes, of the names section; (3)
@@ -64,7 +65,7 @@ static int make_integer (unsigned char *buf)
 
 #define MAGIC 0432
 
-/* In this structure, all char * fields are malloced EXCEPT if the 
+/* In this structure, all char * fields are malloced EXCEPT if the
  * structure is SLTERMCAP.  In that case, only terminal_names is malloced
  * and the other fields are pointers into it.
  */
@@ -76,19 +77,19 @@ typedef struct
 
    unsigned int name_section_size;
    char *terminal_names;
-   
+
    unsigned int boolean_section_size;
    unsigned char *boolean_flags;
-   
+
    unsigned int num_numbers;
    unsigned char *numbers;
-   
+
    unsigned int num_string_offsets;
    unsigned char *string_offsets;
-   
+
    unsigned int string_table_size;
    char *string_table;
-   
+
 } Terminfo_Type;
 
 static char *tcap_getstr (char *, Terminfo_Type *);
@@ -100,10 +101,10 @@ static FILE *open_terminfo (char *file, Terminfo_Type *h)
 {
    FILE *fp;
    unsigned char buf[12];
-   
+
    fp = fopen (file, "rb");
    if (fp == NULL) return NULL;
-   
+
    if ((12 == fread ((char *) buf, 1, 12, fp) && (MAGIC == make_integer (buf))))
      {
 	h->name_section_size = make_integer (buf + 2);
@@ -112,15 +113,15 @@ static FILE *open_terminfo (char *file, Terminfo_Type *h)
 	h->num_string_offsets = make_integer (buf + 8);
 	h->string_table_size = make_integer (buf + 10);
      }
-   else 
+   else
      {
 	fclose (fp);
 	fp = NULL;
      }
    return fp;
 }
-   
-/* 
+
+/*
  * The terminal names section comes next.  It contains the first line of the
  * terminfo description, listing the various names for the terminal,
  * separated by the bar ( | ) character (see term(5)).  The section is
@@ -131,11 +132,11 @@ static FILE *open_terminfo (char *file, Terminfo_Type *h)
 static unsigned char *read_terminfo_section (FILE *fp, unsigned int size)
 {
    char *s;
-   
-   if (NULL == (s = (char *) SLMALLOC (size))) return NULL;
+
+   if (NULL == (s = (char *) SLmalloc (size))) return NULL;
    if (size != fread (s, 1, size, fp))
      {
-	SLFREE (s);
+	SLfree (s);
 	return NULL;
      }
    return (unsigned char *) s;
@@ -160,16 +161,14 @@ static unsigned char *read_boolean_flags (FILE *fp, Terminfo_Type *t)
     * even byte offset. All short integers are aligned on a short word
     * boundary.
     */
-   
+
    unsigned int size = (t->name_section_size + t->boolean_section_size) % 2;
    size += t->boolean_section_size;
-   
+
    return t->boolean_flags = read_terminfo_section (fp, size);
 }
 
-
-
-/* 
+/*
  * The numbers section is similar to the boolean flags section.  Each
  * capability takes up two bytes, and is stored as a short integer.  If the
  * value represented is -1 or -2, the capability is taken to be missing.
@@ -180,7 +179,6 @@ static unsigned char *read_numbers (FILE *fp, Terminfo_Type *t)
    return t->numbers = read_terminfo_section (fp, 2 * t->num_numbers);
 }
 
-
 /* The strings section is also similar.  Each capability is stored as a
  * short integer, in the format above.  A value of -1 or -2 means the
  * capability is missing.  Otherwise, the value is taken as an offset from
@@ -195,7 +193,6 @@ static unsigned char *read_string_offsets (FILE *fp, Terminfo_Type *t)
    return t->string_offsets = (unsigned char *) read_terminfo_section (fp, 2 * t->num_string_offsets);
 }
 
-
 /* The final section is the string table.  It contains all the values of
  * string capabilities referenced in the string section.  Each string is
  * null terminated.
@@ -206,7 +203,6 @@ static char *read_string_table (FILE *fp, Terminfo_Type *t)
    return t->string_table = (char *) read_terminfo_section (fp, t->string_table_size);
 }
 
-
 /*
  * Compiled terminfo(4) descriptions are placed under the directory
  * /usr/share/lib/terminfo.  In order to avoid a linear search of a huge
@@ -217,64 +213,79 @@ static char *read_string_table (FILE *fp, Terminfo_Type *t)
  * are implemented by multiple links to the same compiled file.
  */
 
-#define MAX_TI_DIRS 4
-static char *Terminfo_Dirs [MAX_TI_DIRS] = 
+#define MAX_TI_DIRS 7
+static char *Terminfo_Dirs [MAX_TI_DIRS] =
 {
-   NULL,
+   NULL, /* $HOME/.terminfo */
+   NULL, /* $TERMINFO */
+   "/usr/share/terminfo",
    "/usr/lib/terminfo",
    "/usr/share/lib/terminfo",
+   "/etc/terminfo",
    "/usr/local/lib/terminfo"
 };
 
 char *SLtt_tigetent (char *term)
 {
-   char *tidir; 
+   char *tidir;
    int i;
    FILE *fp = NULL;
-   char file[256];
+   char file[1024];
+   static char home_ti [1024];
+   char *home;
    Terminfo_Type *ti;
 
    if (
-       (term == NULL) 
+       (term == NULL)
 #ifdef SLANG_UNTIC
        && (SLang_Untic_Terminfo_File == NULL)
 #endif
        )
      return NULL;
-   
-   if (NULL == (ti = (Terminfo_Type *) SLMALLOC (sizeof (Terminfo_Type))))
+
+   if (NULL == (ti = (Terminfo_Type *) SLmalloc (sizeof (Terminfo_Type))))
      {
 	return NULL;
      }
-   
+
 #ifdef SLANG_UNTIC
    if (SLang_Untic_Terminfo_File != NULL)
      {
-	fp = open_terminfo (file, ti);
+	fp = open_terminfo (SLang_Untic_Terminfo_File, ti);
 	goto fp_open_label;
      }
    else
 #endif
    /* If we are on a termcap based system, use termcap */
    if (0 == tcap_getent (term, ti)) return (char *) ti;
-       
-   Terminfo_Dirs[0] = getenv ("TERMINFO");
+
+   if (NULL != (home = getenv ("HOME")))
+     {
+	strncpy (home_ti, home, sizeof (home_ti) - 11);
+	home_ti [sizeof(home_ti) - 11] = 0;
+	strcat (home_ti, "/.terminfo");
+	Terminfo_Dirs [0] = home_ti;
+     }
+
+   Terminfo_Dirs[1] = getenv ("TERMINFO");
    i = 0;
    while (i < MAX_TI_DIRS)
      {
 	tidir = Terminfo_Dirs[i];
-	if (tidir != NULL)
+	if ((tidir != NULL)
+	    && (sizeof (file) > strlen (tidir) + 2 + strlen (term)))
 	  {
 	     sprintf (file, "%s/%c/%s", tidir, *term, term);
-	     if (NULL != (fp = open_terminfo (file, ti))) break;
+	     if (NULL != (fp = open_terminfo (file, ti)))
+	       break;
 	  }
 	i++;
      }
 #ifdef SLANG_UNTIC
    fp_open_label:
 #endif
-   
-   if (fp != NULL) 
+
+   if (fp != NULL)
      {
 	if (NULL != read_terminal_names (fp, ti))
 	  {
@@ -291,18 +302,18 @@ char *SLtt_tigetent (char *term)
 				 ti->flags = SLTERMINFO;
 				 return (char *) ti;
 			      }
-			    SLFREE (ti->string_offsets);
+			    SLfree ((char *)ti->string_offsets);
 			 }
-		       SLFREE (ti->numbers);
+		       SLfree ((char *)ti->numbers);
 		    }
-		  SLFREE (ti->boolean_flags);
+		  SLfree ((char *)ti->boolean_flags);
 	       }
-	     SLFREE (ti->terminal_names);
+	     SLfree ((char *)ti->terminal_names);
 	  }
 	fclose (fp);
      }
-   
-   SLFREE (ti);
+
+   SLfree ((char *)ti);
    return NULL;
 }
 
@@ -312,7 +323,7 @@ char *SLtt_tigetent (char *term)
 # define UNTIC_COMMENT(x)
 #endif
 
-typedef struct 
+typedef struct
 {
    char name[3];
    int offset;
@@ -321,17 +332,36 @@ typedef struct
 #endif
 }
 Tgetstr_Map_Type;
-   
-Tgetstr_Map_Type Tgetstr_Map [] =
+
+/* I need to add: K1-5, %0-5(not important), @8, &8... */
+static Tgetstr_Map_Type Tgetstr_Map [] =
 {
-   {"@7", 164 		UNTIC_COMMENT("KEY End")},
-   {"AB", 360 		UNTIC_COMMENT("set a color background")},
-   {"AF", 359 		UNTIC_COMMENT("set a color foreground")},
+   {"#2", 199		UNTIC_COMMENT("Key S-Home")},
+   {"#3", 200		UNTIC_COMMENT("Key S-Insert")},
+   {"#4", 201		UNTIC_COMMENT("Key S-Left")},
+   {"%c", 204		UNTIC_COMMENT("Key S-Next")},
+   {"%e", 206		UNTIC_COMMENT("Key S-Previous")},
+   {"%i", 210		UNTIC_COMMENT("Key S-Right")},
+   {"*4", 191		UNTIC_COMMENT("Key S-Delete")},
+   {"*7", 194		UNTIC_COMMENT("Key S-End")},
+   {"@7", 164 		UNTIC_COMMENT("Key End")},
+   {"AB", 360 		UNTIC_COMMENT("set ANSI color background")},
+   {"AF", 359 		UNTIC_COMMENT("set ANSI color foreground")},
    {"AL", 110 		UNTIC_COMMENT("parm_insert_line")},
    {"DL", 106 		UNTIC_COMMENT("parm_delete_line")},
+   {"F1", 216		UNTIC_COMMENT("key_f11")},
+   {"F2", 217		UNTIC_COMMENT("key_f12")},
+   {"F3", 218		UNTIC_COMMENT("key_f13")},
+   {"F4", 219		UNTIC_COMMENT("key_f14")},
+   {"F5", 220		UNTIC_COMMENT("key_f15")},
+   {"F6", 221		UNTIC_COMMENT("key_f16")},
+   {"F7", 222		UNTIC_COMMENT("key_f17")},
+   {"F8", 223		UNTIC_COMMENT("key_f18")},
+   {"F9", 224		UNTIC_COMMENT("key_f19")},
+   {"FA", 225		UNTIC_COMMENT("key_f20")},
    {"RI", 112 		UNTIC_COMMENT("parm_right_cursor")},
-   {"Sf", 302 		UNTIC_COMMENT("set foreground (color)")},
    {"Sb", 303 		UNTIC_COMMENT("set background (color)")},
+   {"Sf", 302 		UNTIC_COMMENT("set foreground (color)")},
    {"ac", 146 		UNTIC_COMMENT("acs_chars")},
    {"ae", 38 		UNTIC_COMMENT("exit_alt_charset_mode")},
    {"as", 25 		UNTIC_COMMENT("enter_alt_charset_mode")},
@@ -355,6 +385,7 @@ Tgetstr_Map_Type Tgetstr_Map [] =
    {"k7", 73 		UNTIC_COMMENT("key_f7")},
    {"k8", 74 		UNTIC_COMMENT("key_f8")},
    {"k9", 75 		UNTIC_COMMENT("key_f9")},
+   {"k;", 67 		UNTIC_COMMENT("key_f10")},
    {"kA", 78 		UNTIC_COMMENT("key_il")},
    {"kC", 57 		UNTIC_COMMENT("key_clear")},
    {"kD", 59 		UNTIC_COMMENT("key_dc")},
@@ -370,7 +401,6 @@ Tgetstr_Map_Type Tgetstr_Map [] =
    {"kS", 64 		UNTIC_COMMENT("key_eos,")},
    {"kT", 86 		UNTIC_COMMENT("key_stab")},
    {"ka", 56 		UNTIC_COMMENT("key_catab")},
-   {"k;", 67 		UNTIC_COMMENT("key_f10")},
    {"kb", 55 		UNTIC_COMMENT("key_backspace")},
    {"kd", 61 		UNTIC_COMMENT("key_down")},
    {"ke", 88 		UNTIC_COMMENT("End keypad transmit mode")},
@@ -396,22 +426,24 @@ Tgetstr_Map_Type Tgetstr_Map [] =
    {"up", 19 		UNTIC_COMMENT("cursor_up")},
    {"us", 36 		UNTIC_COMMENT("enter_underline_mode")},
    {"vb", 45 		UNTIC_COMMENT("flash_screen")},
+   {"ve", 16 		UNTIC_COMMENT("make cursor very visible")},
    {"vi", 13 		UNTIC_COMMENT("make cursor invisible")},
-   {"vs", 20 		UNTIC_COMMENT("make cursor visible")},
+   {"vs", 20 		UNTIC_COMMENT("make cursor very visible")},
    {"", 0 		UNTIC_COMMENT(NULL)}
 };
 
 static int compute_cap_offset (char *cap, Terminfo_Type *t, Tgetstr_Map_Type *map, unsigned int max_ofs)
 {
    char cha, chb;
-   
+
+   (void) t;
    cha = *cap++; chb = *cap;
-   
+
    while (*map->name != 0)
      {
 	if ((cha == *map->name) && (chb == *(map->name + 1)))
 	  {
-	     if (map->offset >= max_ofs) return -1;
+	     if (map->offset >= (int) max_ofs) return -1;
 	     return map->offset;
 	  }
 	map++;
@@ -419,16 +451,15 @@ static int compute_cap_offset (char *cap, Terminfo_Type *t, Tgetstr_Map_Type *ma
    return -1;
 }
 
-   
 char *SLtt_tigetstr (char *cap, char **pp)
 {
    int offset;
    Terminfo_Type *t;
-   
+
    if ((pp == NULL) || (NULL == (t = (Terminfo_Type *) *pp))) return NULL;
-   
+
    if (t->flags == SLTERMCAP) return tcap_getstr (cap, t);
-   
+
    offset = compute_cap_offset (cap, t, Tgetstr_Map, t->num_string_offsets);
    if (offset < 0) return NULL;
    offset = make_integer (t->string_offsets + 2 * offset);
@@ -436,7 +467,7 @@ char *SLtt_tigetstr (char *cap, char **pp)
    return t->string_table + offset;
 }
 
-Tgetstr_Map_Type Tgetnum_Map[] =
+static Tgetstr_Map_Type Tgetnum_Map[] =
 {
    {"co", 0		UNTIC_COMMENT("columns")},
    {"li", 2		UNTIC_COMMENT("lines")},
@@ -451,7 +482,7 @@ int SLtt_tigetnum (char *cap, char **pp)
 {
    int offset;
    Terminfo_Type *t;
-   
+
    if ((pp == NULL) || (NULL == (t = (Terminfo_Type *) *pp))) return -1;
 
    if (t->flags == SLTERMCAP) return tcap_getnum (cap, t);
@@ -461,7 +492,7 @@ int SLtt_tigetnum (char *cap, char **pp)
    return make_integer (t->numbers + 2 * offset);
 }
 
-Tgetstr_Map_Type Tgetflag_Map[] =
+static Tgetstr_Map_Type Tgetflag_Map[] =
 {
    {"am", 1		UNTIC_COMMENT("auto right margin")},
    {"hs", 9 		UNTIC_COMMENT("has status line")},
@@ -476,19 +507,17 @@ int SLtt_tigetflag (char *cap, char **pp)
 {
    int offset;
    Terminfo_Type *t;
-   
+
    if ((pp == NULL) || (NULL == (t = (Terminfo_Type *) *pp))) return -1;
 
    if (t->flags == SLTERMCAP) return tcap_getflag (cap, t);
-   
+
    offset = compute_cap_offset (cap, t, Tgetflag_Map, t->boolean_section_size);
-   
+
    if (offset < 0) return -1;
    return (int) *(t->boolean_flags + offset);
 }
 
-
-
 /* These are my termcap routines.  They only work with the TERMCAP environment
  * variable.  This variable must contain the termcap entry and NOT the file.
  */
@@ -498,7 +527,7 @@ static int tcap_getflag (char *cap, Terminfo_Type *t)
    char a, b;
    char *f = (char *) t->boolean_flags;
    char *fmax;
-   
+
    if (f == NULL) return 0;
    fmax = f + t->boolean_section_size;
 
@@ -517,10 +546,10 @@ static char *tcap_get_cap (unsigned char *cap, unsigned char *caps, unsigned int
 {
    unsigned char c0, c1;
    unsigned char *caps_max;
-   
+
    c0 = cap[0];
    c1 = cap[1];
-   
+
    if (caps == NULL) return NULL;
    caps_max = caps + len;
    while (caps < caps_max)
@@ -534,7 +563,6 @@ static char *tcap_get_cap (unsigned char *cap, unsigned char *caps, unsigned int
    return NULL;
 }
 
-   
 static int tcap_getnum (char *cap, Terminfo_Type *t)
 {
    cap = tcap_get_cap ((unsigned char *) cap, t->numbers, t->num_numbers);
@@ -562,13 +590,19 @@ static int tcap_getent (char *term, Terminfo_Type *ti)
    unsigned char *buf, *b;
    unsigned char *t;
    int len;
-   
+
    if (SLtt_Try_Termcap == 0) return -1;
-   
+#if 1
+   /* XFREE86 xterm sets the TERMCAP environment variable to an invalid
+    * value.  Specifically, it lacks the tc= string.
+    */
+   if (!strncmp (term, "xterm", 5))
+     return -1;
+#endif
    termcap = (unsigned char *) getenv ("TERMCAP");
    if ((termcap == NULL) || (*termcap == '/')) return -1;
-   
-   /* We have a termcap so lets use it provided it does not have a reference 
+
+   /* We have a termcap so lets use it provided it does not have a reference
     * to another terminal via tc=.  In that case, user terminfo.  The alternative
     * would be to parse the termcap file which I do not want to do right now.
     * Besides, this is a terminfo based system and if the termcap were parsed
@@ -582,37 +616,35 @@ static int tcap_getent (char *term, Terminfo_Type *ti)
 	  return -1;
 	t += (len + 1);
      }
-   
+
    /* malloc some extra space just in case it is needed. */
    len = strlen ((char *) termcap) + 256;
-   if (NULL == (buf = (unsigned char *) SLMALLOC ((unsigned int) len))) return -1;
+   if (NULL == (buf = (unsigned char *) SLmalloc ((unsigned int) len))) return -1;
 
    b = buf;
-   
+
    /* The beginning of the termcap entry contains the names of the entry.
-    * It is terminated by a colon. 
+    * It is terminated by a colon.
     */
-   
+
    ti->terminal_names = (char *) b;
    t = termcap;
    len = tcap_extract_field (t);
    if (len < 0)
      {
-	SLFREE (buf);
+	SLfree ((char *)buf);
 	return -1;
      }
    strncpy ((char *) b, (char *) t, (unsigned int) len);
    b[len] = 0;
    b += len + 1;
    ti->name_section_size = len;
-   
-   
-   /* Now, we are really at the start of the termcap entries.  Point the 
+
+   /* Now, we are really at the start of the termcap entries.  Point the
     * termcap variable here since we want to refer to this a number of times.
     */
    termcap = t + (len + 1);
-   
-   
+
    /* Process strings first. */
    ti->string_table = (char *) b;
    t = termcap;
@@ -620,7 +652,7 @@ static int tcap_getent (char *term, Terminfo_Type *ti)
      {
 	unsigned char *b1;
 	unsigned char *tmax;
-	
+
 	/* We are looking for: XX=something */
 	if ((len < 4) || (t[2] != '=') || (*t == '.'))
 	  {
@@ -629,13 +661,13 @@ static int tcap_getent (char *term, Terminfo_Type *ti)
 	  }
 	tmax = t + len;
 	b1 = b;
-	
+
 	while (t < tmax)
 	  {
 	     ch = *t++;
 	     if ((ch == '\\') && (t < tmax))
 	       {
-		  t = (unsigned char *) SLexpand_escaped_char ((char *) t, (char *) &ch);
+		  t = (unsigned char *) _SLexpand_escaped_char ((char *) t, (char *) &ch);
 	       }
 	     else if ((ch == '^') && (t < tmax))
 	       {
@@ -651,7 +683,7 @@ static int tcap_getent (char *term, Terminfo_Type *ti)
 	b1[2] = (unsigned char) len;    /* replace the = by the length */
 	/* skip colon to next field. */
 	t++;
-     }   
+     }
    ti->string_table_size = (int) (b - (unsigned char *) ti->string_table);
 
    /* Now process the numbers. */
@@ -662,7 +694,7 @@ static int tcap_getent (char *term, Terminfo_Type *ti)
      {
 	unsigned char *b1;
 	unsigned char *tmax;
-	
+
 	/* We are looking for: XX#NUMBER */
 	if ((len < 4) || (t[2] != '#') || (*t == '.'))
 	  {
@@ -671,7 +703,7 @@ static int tcap_getent (char *term, Terminfo_Type *ti)
 	  }
 	tmax = t + len;
 	b1 = b;
-	
+
 	while (t < tmax)
 	  {
 	     *b++ = *t++;
@@ -681,16 +713,16 @@ static int tcap_getent (char *term, Terminfo_Type *ti)
 	len = (int) (b - b1);
 	b1[2] = (unsigned char) len;    /* replace the # by the length */
 	t++;
-     }   
+     }
    ti->num_numbers = (int) (b - ti->numbers);
-   
+
    /* Now process the flags. */
    t = termcap;
    ti->boolean_flags = b;
    while (-1 != (len = tcap_extract_field (t)))
      {
 	/* We are looking for: XX#NUMBER */
-	if ((len != 2) || (*t == '.') || (*t <= ' ')) 
+	if ((len != 2) || (*t == '.') || (*t <= ' '))
 	  {
 	     t += len + 1;
 	     continue;
@@ -699,7 +731,7 @@ static int tcap_getent (char *term, Terminfo_Type *ti)
 	b[1] = t[1];
 	t += 3;
 	b += 2;
-     }   
+     }
    ti->boolean_section_size = (int) (b - ti->boolean_flags);
    ti->flags = SLTERMCAP;
    return 0;
